#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "utils.h"


/*
 * Capacidad definida para buffer de entrada estandar
 */

#define CAPACIDAD 128


/*
 * Lee una linea de texto desde la entrada estandar y retorna el char
 * equivalente; si el texto ingresado no representa un caracter, se le
 * pide al usuario que vuelva a ingresar un caracter. Los espacios que
 * se antepongan o se pongan al final de lo ingresado seran ignorados.
 * Si por alguna razon la linea no se puede leer, la funcion retorna
 * CHAR_MAX
 */

char leer_char(void)
{
    // Intentar leer un caracter ingresado po usuario
    while (true)
    {
        // leer linea de texto, retornando CHAR_MAX en caso de falla
        string line = leer_string();
        if (line == NULL)
            return CHAR_MAX;

        // Retorna un char, si y solo si se ingreso un char.
        char c1, c2;
        if (sscanf(line, " %c %c", &c1, &c2) == 1)
        {
            free(line);
            return c1;
        }
        else
        {
            free(line);
            printf("Ingrese nuevamente: ");
        }
    }
}


/*
 * Lee una linea de texto desde la entrada estandar y retorna un double
 * equivalente al texto ingresado; Si el texto ingresado no representa
 * un double, se le pide al usuario que reingrese un valor. Si se
 * anteponen, o se ponen espacios al final del numero ingresado, estos
 * seran ignorados. Esta funcion no es capaz de detectar overflow y
 * underflow. Si una linea no puede leerse, se retornara DBL_MAX.
 */

double leer_double(void)
{
    while (true)
    {
        string line = leer_string();
        if (line == NULL)
            return DBL_MAX;

        double d; char c;
        if (sscanf(line, " %lf %c", &d, &c) == 1)
        {
            free(line);
            return d;
        }
        else
        {
            free(line);
            printf("Ingrese nuevamente: ");
        }
    }
}


/*
 * Lee una linea de texto desde la entrada estandar y retorna un float
 * equivalente al texto ingresado; Si el texto ingresado no representa
 * un flotante, se le pide al usuario que reingrese un valor. Si se
 * anteponen, o se ponen espacios al final del numero ingresado, estos
 * seran ignorados. Esta funcion no es capaz de detectar overflow y
 * underflow. Si una linea no puede leerse, se retornara FLT_MAX.
 */

float leer_float(void)
{
    while (true)
    {
        string line = leer_string();
        if (line == NULL)
            return FLT_MAX;

        char c; float f;
        if (sscanf(line, " %f %c", &f, &c) == 1)
        {
            free(line);
            return f;
        }
        else
        {
            free(line);
            printf("Ingrese nuevamente: ");
        }
    }
}


/*
 * Lee una linea desde la entrada estandar y la retorna como un
 * int en el rango: [-2^31 + 1, 2^31 - 2]; Si el texto no representa
 * un entero, se le pedira al usuario que reingrese un valor. Los espacios
 * al inicio y final del texto ingresado seran ignorados. Esta funcion no
 * detexta overflow. Si no se puede leer la linea de texto, se retornara INT_MAX
 */

int leer_int(void)
{
    while (true)
    {
        string line = leer_string();
        if (line == NULL)
            return INT_MAX;

        int n; char c;
        if (sscanf(line, " %d %c", &n, &c) == 1)
        {
            free(line);
            return n;
        }
        else
        {
            free(line);
            printf("Ingrese nuevamente: ");
        }
    }
}


/*
 * Lee una linea desde la entrada estandar y la retorna como un long
 * long en el rango: [-2^63 + 1, 2^63 - 2]; Si el texto no representa
 * un long long, se le pedira al usuario que reingrese un valor. Los espacios
 * al inicio y final del texto ingresado seran ignorados. Esta funcion no
 * detecta overflow. Si no se puede leer la linea de texto, se retornara LLONG_MAX
 */

long long leer_long_long(void)
{
    while (true)
    {
        string line = leer_string();
        if (line == NULL)
            return LLONG_MAX;

        long long n; char c;
        if (sscanf(line, " %lld %c", &n, &c) == 1)
        {
            free(line);
            return n;
        }
        else
        {
            free(line);
            printf("Ingrese nuevamente: ");
        }
    }
}


/* Lee una linea de texto desde la entrada estandar y la retorna
 * como un char *, eliminando los '\n' (es decir, si el usuario
 * ingresa solo '\n', se retornara "", no NULL). Retorna NULL en
 * caso de error o si no se pudo leer linea (EOF por ejemplo). Los
 * espacios al inicio y al final de la entrada no son ignorados.
 * El string leido se guarda en el heap (a traves de malloc), por
 * lo que la memoria debera ser liberada mediante free para evitar
 * fugas de memoria.
 */

string leer_string(void)
{
    // buffer de tamao dinamico para almacenar chars
    string buffer = NULL;

    // capacidad del buffer
    unsigned int capacidad = 0;

    // cantidad de caracteres actualmente en el buffer
    unsigned int n = 0;

    // char leido o EOF
    int c;

    // Lee chars de forma iterativa desde la entrada estandar
    while ((c = fgetc(stdin)) != '\n' && c != EOF)
    {
        // Aumentar capacidad del buffer en caso de ser necesario
        if (n + 1 > capacidad)
        {
            // Empezar con la capacidad definida CAPACIDAD, si no, duplicar
            if (capacidad == 0)
                capacidad = CAPACIDAD;
            else if (capacidad <= (UINT_MAX / 2))
                capacidad *= 2;
            else
            {
                free(buffer);
                return NULL;
            }

            // extender la capacidad del buffer
            string temp = realloc(buffer, capacidad * sizeof(char));
            if (temp == NULL)
            {
                free(buffer);
                return NULL;
            }
            buffer = temp;
        }

        // Agregar el char leido al buffer
        buffer[n++] = c;
    }

    // Retorna NULL si el usuario no ingreso nada
    if (n == 0 && c == EOF)
        return NULL;

    // Minimizar el buffer a la cantidad de caracteres ingresados
    string minimal = malloc((n + 1) * sizeof(char));
    strncpy(minimal, buffer, n);
    free(buffer);

    // Agregar caracter nulo al final del string
    minimal[n] = '\0';

    // retornar string
    return minimal;
}


